from typing import Any, Optional

import click
from ape.logging import logger
from eth_account import Account as EthAccount
from eth_account.signers.local import LocalAccount
from eth_utils import to_bytes


def get_eth_account(private_key: str) -> Optional[LocalAccount]:
    try:
        return EthAccount.from_key(to_bytes(hexstr=private_key))
    except Exception as err:
        logger.log_error(err)
        return None


def agree_to_sign(message: Any, message_type_name: str) -> bool:
    return click.confirm(f"Sign {message_type_name}: {message}")



import tempfile
from pathlib import Path

import ape
import keyring.backend
import pytest
from ape._cli import cli as root_cli
from click.testing import CliRunner

from ape_keyring._secrets import get_secret_manager
from ape_keyring.storage import SecretStorage
from ape_keyring.testing import EphemeralBackend

ape.config.DATA_FOLDER = Path(tempfile.mkdtemp()).resolve()
TEMP_DATA_FOLDER = Path(tempfile.mkdtemp()).resolve()
PROJECT_DIRECTORY = Path(__file__).parent


@pytest.fixture(scope="session", autouse=True)
def from_tests_directory(config):
    with config.using_project(PROJECT_DIRECTORY):
        yield


@pytest.fixture(scope="session", autouse=True)
def backend(accounts):
    backend = EphemeralBackend()
    keyring.set_keyring(backend)
    return backend


@pytest.fixture(scope="session", autouse=True)
def storage():
    return SecretStorage("secrets")


@pytest.fixture(scope="session", autouse=True)
def secret_manager(container, storage):
    return get_secret_manager(PROJECT_DIRECTORY, storage=storage)


@pytest.fixture(scope="session", autouse=True)
def container(accounts, storage):
    container = accounts.containers["keyring"]
    container.storage = storage
    return container


@pytest.fixture(scope="session")
def accounts():
    return ape.accounts


@pytest.fixture(scope="session")
def config():
    return ape.config


@pytest.fixture
def runner():
    return CliRunner()


@pytest.fixture(scope="session")
def cli():
    return root_cli


@pytest.fixture(scope="session")
def existing_alias():
    return "existing-test-alias"


@pytest.fixture(scope="session")
def non_existing_alias():
    return "non_existing-test-alias"


@pytest.fixture
def test_account(accounts):
    return accounts.test_accounts[0]


@pytest.fixture
def address(test_account):
    return test_account.address


@pytest.fixture
def private_key(test_account):
    return test_account.private_key


@pytest.fixture
def keyring_account(container, private_key, existing_alias):
    if existing_alias not in container.aliases:
        container.create_account(existing_alias, private_key)

    yield container.load(existing_alias)

    if existing_alias in container.aliases:
        container.delete_account(existing_alias)



import click

from ape_keyring._cli.accounts import account_cli
from ape_keyring._cli.secrets import secrets


@click.group("keyring")
def cli():
    """Manage accounts and secrets"""
    pass


cli.add_command(account_cli)
cli.add_command(secrets)



import os
from enum import Enum
from pathlib import Path
from typing import List, Optional, Union

from ape.utils import ManagerAccessMixin, load_config

from ape_keyring.config import KeyringConfig
from ape_keyring.storage import SecretStorage, secret_storage


class Scope(Enum):
    GLOBAL = "global"
    PROJECT = "project"


class SecretManager(ManagerAccessMixin):
    def __init__(self, project_path: Path, storage: SecretStorage):
        self._path = project_path
        self._storage = storage

    @property
    def project_name(self) -> str:
        return self._path.stem

    @property
    def _project_key_prefix(self) -> str:
        return "<<project="

    @property
    def _project_key(self):
        return f"{self._project_key_prefix}{self.project_name}>>"

    @property
    def secrets_exist(self) -> bool:
        return self.project_keys or self.global_keys  # type: ignore

    @property
    def project_keys(self) -> List[str]:
        keys = self._storage.keys
        return [k.replace(self._project_key, "") for k in keys if self._project_key in k]

    @property
    def global_keys(self) -> List[str]:
        keys = self._storage.keys
        return [k for k in keys if k not in self.project_keys and self._project_key_prefix not in k]

    @property
    def config(self) -> KeyringConfig:
        raw_config = load_config(self._path / "ape-config.yaml")
        return KeyringConfig.parse_obj(raw_config.get("keyring", {}))

    @property
    def do_set_env_vars(self) -> bool:
        stored_value = self.config.dict().get("set_env_vars") or "f"
        return str(stored_value).lower() in ["1", "true", "t"]

    def get_secret(self, key, scope: Union[str, Scope] = Scope.GLOBAL) -> str:
        key = self._get_key(key, scope)
        return self._storage.get_secret(key) or ""

    def store_secret(self, key: str, secret: str, scope: Union[str, Scope] = Scope.GLOBAL):
        key = self._get_key(key, scope)
        self._storage.store_secret(key, secret)
        if self.do_set_env_vars:
            self._set_env_var(key, secret)

    def delete_secret(self, key: str, scope: Union[str, Scope] = Scope.GLOBAL) -> bool:
        key = self._get_key(key, scope)
        did_delete = self._storage.delete_secret(key)
        if self.do_set_env_vars:
            self._unset_env_var(key)

        return did_delete

    def set_environment_variables(self):
        """
        Set the environment variables if told to from the config.
        """
        if not self.do_set_env_vars:
            return

        for key, value in self._storage:
            self._set_env_var(key, value)

    def _get_key(self, key: str, scope: Union[str, Scope]):
        return f"{key}{self._project_key}" if scope in [Scope.PROJECT, Scope.PROJECT.value] else key

    def _set_env_var(self, key: str, value: str):
        # Strip of 'project=' and '<<>>' parts before setting as env var.
        key = self._extract_env_var_key(key)
        os.environ[key] = value

    def _unset_env_var(self, key: str):
        key = self._extract_env_var_key(key)
        if key in os.environ:
            del os.environ[key]

    def _extract_env_var_key(self, key: str):
        return key.split(self._project_key)[0]


def get_secret_manager(project_path: Path, storage: Optional[SecretStorage] = None):
    return SecretManager(project_path, storage or secret_storage)


__all__ = ["get_secret_manager", "Scope", "SecretManager"]



import json
from pathlib import Path
from typing import Any, Dict, List, Optional

import keyring
from ape.logging import logger
from ape.utils import ManagerAccessMixin
from keyring.errors import PasswordDeleteError

SERVICE_NAME = "ape-keyring"
ACCOUNTS_TRACKER_KEY = "ape-keyring-accounts"
SECRETS_TRACKER_KEY = "ape-keyring-secrets"


class SecretStorage(ManagerAccessMixin):
    def __init__(self, tracker_key: str):
        """
        Initialize a new base-storage class.

        Args:
            tracker_key (str): The key-name for storing a comma-separated list
              of items tracked.
        """

        self._tracker_key = tracker_key

    def __iter__(self):
        for key in self.keys:
            secret = self.get_secret(key)
            if secret:
                yield key, secret

    @property
    def data_folder(self) -> Path:
        try:
            return self.account_manager.containers["keyring"].data_folder
        except ValueError as err:
            # TODO: Come up with better fix
            if str(err) == "Value not set. Please inject this property before calling.":
                # Fixes race condition when using set_env_vars: True
                # and plugins are not loaded properly yet.
                return Path.home() / ".ape" / "keyring"

            raise  # Original error

    @property
    def data_file_path(self) -> Path:
        return self.data_folder / "data.json"

    @property
    def plugin_data(self) -> Dict:
        if self.data_file_path.is_file():
            return json.loads(self.data_file_path.read_text())

        return {}

    @property
    def keys(self) -> List[str]:
        """
        A list of stored item keys. Each key can unlock a secret.

        Returns:
            List[str]
        """

        return self.plugin_data.get(self._tracker_key, [])

    def get_secret(self, key: str) -> Optional[str]:
        """
        Get a secret.

        Args:
            key (str): The key for the secret, such as an account alias
              or environment variable name.

        Returns:
            str: The secret value from the OS secure-storage.
        """
        if key not in self.keys:
            return None

        return _get_secret(key)

    def store_secret(self, key: str, secret: str):
        """
        Add a new item to be tracked.

        Args:
            key (str): The new key for the item.
            secret (str): The value of the secret to store.
        """

        if key not in self.keys:
            self._track([*self.keys, key])

        _set_secret(key, secret)

    def delete_secret(self, key: str):
        if key in self.keys:
            self._track([k for k in self.keys if k != key])

        return _delete_secret(key)

    def delete_all(self):
        for key in self.keys:
            _delete_secret(key)

    def _track(self, new_keys: List[str]):
        self._store_public_data(self._tracker_key, new_keys)

    def _store_public_data(self, key: str, value: Any):
        self.data_folder.mkdir(exist_ok=True, parents=True)
        data = {**dict(self.plugin_data), key: value}
        if self.data_file_path.exists():
            self.data_file_path.unlink()

        self.data_file_path.write_text(json.dumps(data))


account_storage = SecretStorage(ACCOUNTS_TRACKER_KEY)
"""A storage class for storing secrets."""


secret_storage = SecretStorage(SECRETS_TRACKER_KEY)
"""A storage class for storing secrets."""


def _get_secret(key: str) -> Optional[str]:
    try:
        return keyring.get_password(SERVICE_NAME, key)
    except KeyError:
        return None


def _set_secret(key: str, secret: str):
    if not key or not secret:
        return

    keyring.set_password(SERVICE_NAME, key, secret)


def _delete_secret(key: str):
    if not key:
        return False

    try:
        keyring.delete_password(SERVICE_NAME, key)
        return True
    except (PasswordDeleteError, AssertionError) as err:
        logger.debug(err)
        return False



import click

from ape_keyring._secrets import Scope


def secret_argument(callback=None):
    return click.argument("secret", callback=callback)


def scope_option():
    return click.option(
        "--scope",
        help="Set to 'project' to limit the scope to the current project.",
        default="global",
        callback=lambda ctx, param, value: Scope(value),
    )
